interface Schema {
    [K: string]: Schema | string;
}
type InferLiteralString<T, TD extends string> = T extends infer K ? K extends string ? K extends '' ? `${TD}` : `${K}` : `${TD}` : `${TD}`;
type ScopePrefixer<T, S> = T extends infer K ? K extends string ? K extends '' ? '' : `${K}${InferLiteralString<S, '.'>}` : never : never;
type UnionScopes<T, Prefix extends string = '', Spacer extends string = '.'> = T extends Record<string, any> ? {
    [K in keyof T]: K extends string ? T[K] extends Record<string, any> ? UnionScopes<T[K], `${ScopePrefixer<Prefix, Spacer>}${K}`, InferLiteralString<Spacer, '.'>> : `${ScopePrefixer<Prefix, Spacer>}${K}` : never;
}[keyof T] : Prefix;
type PermitNode = {
    scope: string;
    granted: boolean;
    desc: string;
    toggle: () => void;
};
type TransformSchema<T, TValue> = T extends Schema ? {
    [P in keyof T]: T[P] extends Schema ? TransformSchema<T[P], TValue> : TValue;
} : TValue;
type SchemaOptions = {
    /** add prefix to scope. default to '' (none) */
    prefix: string;
    /** string to use for spacer for scope. default to '.' */
    spacer: string;
};

type SchemaNode = {
    scope: string;
    description: string;
};
declare class AccessSchema<T extends Schema = Schema> {
    private _options;
    private _schema;
    private _scopes;
    constructor(schema: T, options?: Partial<SchemaOptions>);
    private createSchemaNode;
    private resetSchemaNode;
    private get scopesList();
    /** utility function for finding scope */
    private _get;
    /** readonly schema options */
    get options(): SchemaOptions;
    /** readonly AccessNodes  */
    get nodes(): TransformSchema<T, string>;
    /** export schema to list of scope and description */
    export(): SchemaNode[];
    /** get schema node by string scope */
    get(scope: string): SchemaNode | undefined;
    /** utility function for checking if there is a schema node that have an exact match with given scope */
    has(scope: string): boolean;
    /** find schema node that has the given substring in scope or description. */
    search(substring: string, fn?: (n: SchemaNode) => boolean): SchemaNode[];
    /** find all child of the given string scope */
    getChild(scope: string): SchemaNode[];
    /** Map over schema and transform leaf node by given handler */
    map<V>(handler: (node: SchemaNode) => V): TransformSchema<T, V>;
    /**
     * Create permit from schema with given granted scopes
     */
    permit(scopes: string[] | SchemaNode[]): {
        get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<T, Prefix, Spacer>) => PermitNode | undefined;
        has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<T, Prefix_1, Spacer_1>) => boolean;
        search: (substring: string, handle: (node: PermitNode) => void) => void;
        getChild: (scope: string) => PermitNode[];
        reset: () => void;
        grant: (...scopes: string[] | SchemaNode[]) => void;
        deny: (...scopes: string[] | SchemaNode[]) => void;
        check(scope: string): {
            readonly exact: boolean;
            readonly hasChild: boolean;
            readonly hasAnyChildGranted: boolean;
            readonly hasAllChildGranted: boolean;
            readonly hasAllChildDenied: boolean;
        };
        readonly export: {
            readonly all: {
                scope: string;
                desc: string;
            }[];
            readonly granted: {
                scope: string;
                desc: string;
            }[];
            readonly denied: {
                scope: string;
                desc: string;
            }[];
        };
        readonly granted: {
            scope: string;
            desc: string;
        }[];
        readonly denied: {
            scope: string;
            desc: string;
        }[];
        readonly nodes: TransformSchema<T, PermitNode>;
        readonly can: TransformSchema<T, boolean>;
    };
}

type PermitScope = Omit<PermitNode, 'toggle' | 'granted'>;
type Permit<T extends Schema> = ReturnType<typeof permit$1<T>>;
declare function permit$1<T extends Schema>(access: AccessSchema<T>, values: string[]): {
    get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<T, Prefix, Spacer>) => PermitNode | undefined;
    has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<T, Prefix_1, Spacer_1>) => boolean;
    search: (substring: string, handle: (node: PermitNode) => void) => void;
    getChild: (scope: string) => PermitNode[];
    reset: () => void;
    grant: (...scopes: string[] | SchemaNode[]) => void;
    deny: (...scopes: string[] | SchemaNode[]) => void;
    /** utility for checking string scope permit value */
    check(scope: string): {
        readonly exact: boolean;
        readonly hasChild: boolean;
        readonly hasAnyChildGranted: boolean;
        readonly hasAllChildGranted: boolean;
        readonly hasAllChildDenied: boolean;
    };
    /** getter for exporting scopes */
    readonly export: {
        readonly all: PermitScope[];
        readonly granted: PermitScope[];
        readonly denied: PermitScope[];
    };
    /** readonly enabled access node */
    readonly granted: PermitScope[];
    /** readonly disabled access node */
    readonly denied: PermitScope[];
    /** readonly AccessNodes  */
    readonly nodes: TransformSchema<T, PermitNode>;
    /** readonly assertion utility for granted scopes  */
    readonly can: TransformSchema<T, boolean>;
};

/**
 * Create access control schema
 *
 * @param {T} schema
 * @param {Partial<SchemaOptions>} [options={}]
 * @return {AccessSchema} access schema
 */
declare function createSchema<T extends Schema>(schema: T, options?: Partial<SchemaOptions>): AccessSchema<T>;
type PermitOptions = {
    granted: string[] | SchemaNode[];
    denied: string[] | SchemaNode[];
} & SchemaOptions;
/**
 * Create user permit
 *
 * @export
 * @template T
 * @param {AccessSchema<T>} schema
 * @param {(string[] | SchemaNode[])} scopes
 * @return {Permit}
 */
declare function createPermit<T extends Schema>(schema: T, options?: Partial<PermitOptions>): {
    get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<T, Prefix, Spacer>) => PermitNode | undefined;
    has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<T, Prefix_1, Spacer_1>) => boolean;
    search: (substring: string, handle: (node: PermitNode) => void) => void;
    getChild: (scope: string) => PermitNode[];
    reset: () => void;
    grant: (...scopes: string[] | SchemaNode[]) => void;
    deny: (...scopes: string[] | SchemaNode[]) => void;
    check(scope: string): {
        readonly exact: boolean;
        readonly hasChild: boolean;
        readonly hasAnyChildGranted: boolean;
        readonly hasAllChildGranted: boolean;
        readonly hasAllChildDenied: boolean;
    };
    readonly export: {
        readonly all: {
            scope: string;
            desc: string;
        }[];
        readonly granted: {
            scope: string;
            desc: string;
        }[];
        readonly denied: {
            scope: string;
            desc: string;
        }[];
    };
    readonly granted: {
        scope: string;
        desc: string;
    }[];
    readonly denied: {
        scope: string;
        desc: string;
    }[];
    readonly nodes: TransformSchema<T, PermitNode>;
    readonly can: TransformSchema<T, boolean>;
};
type SchemaType<T> = T extends AccessSchema<infer S> ? S : never;
/**
 * Create guard utility that check based on given guard function
 *
 * @export
 * @template T
 * @template P
 * @param {AccessSchema<T>} schema
 * @param {(permit: Permit<T, P>) => boolean} guard
 * @return {*}
 */
declare function createGuard<T extends Schema>(schema: AccessSchema<T>, guard: (permit: Permit<T>) => boolean): {
    grant(scopes: string[]): void;
    deny(scopes: string[]): void;
    readonly isAllowed: boolean;
    readonly isDenied: boolean;
};

type _permit_AccessSchema<T extends Schema = Schema> = AccessSchema<T>;
declare const _permit_AccessSchema: typeof AccessSchema;
type _permit_Permit<T extends Schema> = Permit<T>;
type _permit_Schema = Schema;
type _permit_SchemaNode = SchemaNode;
type _permit_SchemaOptions = SchemaOptions;
type _permit_SchemaType<T> = SchemaType<T>;
type _permit_UnionScopes<T, Prefix extends string = '', Spacer extends string = '.'> = UnionScopes<T, Prefix, Spacer>;
declare const _permit_createGuard: typeof createGuard;
declare const _permit_createPermit: typeof createPermit;
declare const _permit_createSchema: typeof createSchema;
declare namespace _permit {
  export {
    _permit_AccessSchema as AccessSchema,
    _permit_Permit as Permit,
    _permit_Schema as Schema,
    _permit_SchemaNode as SchemaNode,
    _permit_SchemaOptions as SchemaOptions,
    _permit_SchemaType as SchemaType,
    _permit_UnionScopes as UnionScopes,
    _permit_createGuard as createGuard,
    _permit_createPermit as createPermit,
    _permit_createSchema as createSchema,
  };
}

type ZodType<T> = {
    parse: (data: any) => T;
    array: () => ZodType<T[]>;
};
type parseOptions<T> = {
    default?: T;
    parseHandler?: (value: DataType) => T;
};
type DataRecord = {
    [P in string]: DataType;
};
type DataType = DataRecord | DataRecord[] | any[] | string | number | boolean | null | undefined;
type Handler<T> = (row: Inferer) => T;
type Inferer = ReturnType<typeof infer>;
declare function infer(data: DataType): {
    map: <T>(path: string | undefined, handler: Handler<T>) => T[];
    array: <T_1 = DataType>(path: string, schema: ZodType<T_1> | Handler<T_1>, options?: parseOptions<T_1[]>) => T_1[];
    object: <T_2>(path: string, schema: ZodType<T_2> | Handler<T_2>, options?: parseOptions<T_2>) => T_2;
    string: (path?: string, options?: parseOptions<string>) => string;
    number: (path?: string, options?: parseOptions<number> & {
        float?: boolean;
    }) => number;
    boolean: (path?: string, options?: parseOptions<boolean>) => boolean;
    nullable: (path?: string) => {
        string: (options?: parseOptions<string>) => string | null;
        number: (options?: parseOptions<number> & {
            float?: boolean;
        }) => number | null;
        boolean: (options?: parseOptions<boolean>) => boolean | null;
        stringOrNumber: (options?: parseOptions<string | number> & {
            float?: boolean;
        }) => string | number | null;
        array: <T_3 = DataType>(handler: ZodType<T_3> | Handler<T_3>, options?: parseOptions<T_3[]>) => T_3[] | null;
        object: <T_4>(handler: ZodType<T_4> | Handler<T_4>, options?: parseOptions<T_4>) => T_4 | null;
        map: <T_5>(handler: Handler<T_5>) => T_5[] | null;
    };
    get: (path?: string) => DataType;
    stringOrNumber: (path?: string, options?: parseOptions<string | number> & {
        float?: boolean;
    }) => string | number;
    readonly raw: DataType;
    has(path: string): boolean;
};

type Ctx = Record<string, any>;
type ContextPipe<T extends Ctx, V> = T & {
    value: V;
};
type Procedure<T extends Ctx, I extends any[], O> = ((...input: I) => O) & {
    zod: <V>(schema: {
        parse: (data: any) => V;
    }) => Procedure<T, I, V>;
    zodSafe: <V>(schema: {
        safeParse: (data: any) => V;
    }) => Procedure<T, I, V>;
    zodAsync: <V>(schema: {
        parseAsync: (data: any) => V;
    }) => Procedure<T, I, V>;
    zodSafeAsync: <V>(schema: {
        safeParseAsync: (data: any) => V;
    }) => Procedure<T, I, V>;
    infer: <V>(inferFn: (inferer: Inferer) => V) => Procedure<T, I, V>;
    pipe: <V>(pipeFn: (context: ContextPipe<T, O>, ...input: I) => V) => Procedure<T, I, V>;
};
type ProcedureOptions = {
    onAfterCalled: (value: any) => void;
    onBeforeCalled: (input: any[]) => void;
};
type Context<T extends Record<string, any> = any> = (() => T) & {
    context: () => T;
    procedure: <I extends any[], O>(fn: (context: T, ...input: I) => O) => Procedure<T, I, O>;
};
type ContextOptions = {
    immidiate: boolean;
    onAfterCalled: (value: any) => void;
    onBeforeCalled: (input: any[]) => void;
};
type InferContext<C extends Context> = C extends Context<infer T> ? T : never;
type Pipes<I extends any[], O> = ((...input: I) => O) & {
    zod: <V>(schema: {
        parse: (data: any) => V;
    }) => Pipes<I, V>;
    zodSafe: <V>(schema: {
        safeParse: (data: any) => V;
    }) => Pipes<I, V>;
    zodAsync: <V>(schema: {
        parseAsync: (data: any) => V;
    }) => Pipes<I, V>;
    zodSafeAsync: <V>(schema: {
        safeParseAsync: (data: any) => V;
    }) => Pipes<I, V>;
    infer: <V>(inferFn: (inferer: Inferer) => V) => Pipes<I, V>;
    pipe: <V>(pipeFn: (value: O, ...input: I) => V) => Pipes<I, V>;
};
type PipesOptions = {
    onAfterCalled: (value: any) => void;
    onBeforeCalled: (input: any[]) => void;
};
declare function createProcedure<T extends Record<string, any>, I extends any[], O>(context: () => T, fn: (context: T, ...input: I) => O, options?: Partial<ProcedureOptions>): Procedure<T, I, O>;
declare function createContext<T extends Record<string, any>>(context: () => T, options?: Partial<ContextOptions>): {
    (): T;
    context: () => T;
    procedure<I extends any[], O>(fn: (context: T, ...input: I) => O): Procedure<T, I, O>;
};
declare function createPipe<I extends any[], O>(fn: (...input: I) => O, options?: Partial<ProcedureOptions>): Pipes<I, O>;

type _garda_Context<T extends Record<string, any> = any> = Context<T>;
type _garda_ContextOptions = ContextOptions;
type _garda_InferContext<C extends Context> = InferContext<C>;
type _garda_Pipes<I extends any[], O> = Pipes<I, O>;
type _garda_PipesOptions = PipesOptions;
type _garda_Procedure<T extends Ctx, I extends any[], O> = Procedure<T, I, O>;
type _garda_ProcedureOptions = ProcedureOptions;
declare const _garda_createContext: typeof createContext;
declare const _garda_createPipe: typeof createPipe;
declare const _garda_createProcedure: typeof createProcedure;
declare namespace _garda {
  export {
    _garda_Context as Context,
    _garda_ContextOptions as ContextOptions,
    _garda_InferContext as InferContext,
    _garda_Pipes as Pipes,
    _garda_PipesOptions as PipesOptions,
    _garda_Procedure as Procedure,
    _garda_ProcedureOptions as ProcedureOptions,
    _garda_createContext as createContext,
    _garda_createPipe as createPipe,
    _garda_createProcedure as createProcedure,
  };
}

declare const garda: typeof _garda;
declare const permit: typeof _permit;

export { DataRecord, DataType, Inferer, ZodType, garda, infer, permit };
