/* eslint-disable @typescript-eslint/no-empty-function */
/* eslint-disable @typescript-eslint/no-explicit-any */

import { DataType, Inferer, infer } from '../infer';

type Ctx = Record<string, any>;
type ContextPipe<T extends Ctx, V> = T & { value: V };
type PipeInjector<T extends Ctx, V> = { context: T; value: V };
export type Procedure<T extends Ctx, I extends any[], O> = ((
  ...input: I
) => O) & {
  zod: <V>(schema: { parse: (data: any) => V }) => Procedure<T, I, V>;
  zodSafe: <V>(schema: { safeParse: (data: any) => V }) => Procedure<T, I, V>;
  zodAsync: <V>(schema: { parseAsync: (data: any) => V }) => Procedure<T, I, V>;
  zodSafeAsync: <V>(schema: {
    safeParseAsync: (data: any) => V;
  }) => Procedure<T, I, V>;
  infer: <V>(inferFn: (inferer: Inferer) => V) => Procedure<T, I, V>;
  pipe: <V>(
    pipeFn: (context: ContextPipe<T, O>, ...input: I) => V
  ) => Procedure<T, I, V>;
};

export type ProcedureOptions = {
  onAfterCalled: (value: any) => void;
  onBeforeCalled: (input: any[]) => void;
};

export type Context<T extends Record<string, any> = any> = (() => T) & {
  context: () => T;
  procedure: <I extends any[], O>(
    fn: (context: T, ...input: I) => O
  ) => Procedure<T, I, O>;
};

export type ContextOptions = {
  immidiate: boolean;
  onAfterCalled: (value: any) => void;
  onBeforeCalled: (input: any[]) => void;
};

export type InferContext<C extends Context> = C extends Context<infer T>
  ? T
  : never;

export type Pipes<I extends any[], O> = ((...input: I) => O) & {
  zod: <V>(schema: { parse: (data: any) => V }) => Pipes<I, V>;
  zodSafe: <V>(schema: { safeParse: (data: any) => V }) => Pipes<I, V>;
  zodAsync: <V>(schema: { parseAsync: (data: any) => V }) => Pipes<I, V>;
  zodSafeAsync: <V>(schema: {
    safeParseAsync: (data: any) => V;
  }) => Pipes<I, V>;
  infer: <V>(inferFn: (inferer: Inferer) => V) => Pipes<I, V>;
  pipe: <V>(pipeFn: (value: O, ...input: I) => V) => Pipes<I, V>;
};

export type PipesOptions = {
  onAfterCalled: (value: any) => void;
  onBeforeCalled: (input: any[]) => void;
};

export function createProcedure<
  T extends Record<string, any>,
  I extends any[],
  O
>(
  context: () => T,
  fn: (context: T, ...input: I) => O,
  options: Partial<ProcedureOptions> = {}
) {
  // recursive pipe generator
  const generatePipe = <Vprev, Vnext>(
    proc: (...input: I) => PipeInjector<T, Vprev>,
    pipe: (context: ContextPipe<T, Vprev>, ...input: I) => Vnext
  ): Procedure<T, I, Vnext> => {
    // pipe injector for context and previous value
    const injector = (...input: I) => {
      const pipeReturn = proc(...input);
      const value = pipe(
        { ...pipeReturn.context, value: pipeReturn.value },
        ...input
      );
      return { context: pipeReturn.context, value };
    };

    // create pipe caller
    const caller = (...input: I) => {
      const { context, value } = proc(...input);
      const finalValue = pipe({ ...context, value }, ...input);

      if (typeof options.onAfterCalled === 'function') {
        options.onAfterCalled(finalValue);
      }

      return finalValue;
    };

    // generate pipe function recursively
    caller.pipe = <V>(
      pipeFn: (context: ContextPipe<T, Vnext>, ...input: I) => V
    ) => {
      return generatePipe<Vnext, V>(injector, pipeFn) as Procedure<T, I, V>;
    };

    // create infer function
    caller.infer = <V>(inferFn: (inferer: Inferer) => V) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, ({ value }, ..._: I) => {
        return inferFn(infer(value as DataType));
      }) as Procedure<T, I, V>;
    };

    caller.zod = <V>(schema: { parse: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, ({ value }, ..._: I) => {
        return schema.parse(value);
      }) as Procedure<T, I, V>;
    };

    caller.zodSafe = <V>(schema: { safeParse: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, ({ value }, ..._: I) => {
        return schema.safeParse(value);
      }) as Procedure<T, I, V>;
    };

    caller.zodAsync = <V>(schema: { parseAsync: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, ({ value }, ..._: I) => {
        return schema.parseAsync(value);
      }) as Procedure<T, I, V>;
    };

    caller.zodSafeAsync = <V>(schema: { safeParseAsync: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, ({ value }, ..._: I) => {
        return schema.safeParseAsync(value);
      }) as Procedure<T, I, V>;
    };

    // return as new procedure
    return caller as Procedure<T, I, Vnext>;
  };

  // prepare caller
  const procedure = (...input: I) => {
    const ctx = context();
    if (typeof options.onBeforeCalled === 'function') {
      options.onBeforeCalled(input);
    }

    const value = fn(ctx, ...input);

    if (typeof options.onAfterCalled === 'function') {
      options.onAfterCalled(value);
    }

    return value;
  };

  // create initial injector
  const injector = (...input: I) => {
    const ctx = context();
    if (typeof options.onBeforeCalled === 'function') {
      options.onBeforeCalled(input);
    }

    return {
      context: ctx,
      value: fn(ctx, ...input),
    };
  };

  // create first pipe function
  procedure.pipe = <V>(
    pipeFn: (context: ContextPipe<T, O>, ...input: I) => V
  ) => {
    return generatePipe<O, V>(injector, pipeFn) as Procedure<T, I, V>;
  };

  // create first infer function
  procedure.infer = <V>(inferFn: (inferer: Inferer) => V) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, ({ value }, ..._: I) => {
      return inferFn(infer(value as DataType));
    }) as Procedure<T, I, V>;
  };

  procedure.zod = <V>(schema: { parse: (data: any) => V }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, ({ value }, ..._: I) => {
      return schema.parse(value);
    }) as Procedure<T, I, V>;
  };

  procedure.zodSafe = <V>(schema: { safeParse: (data: any) => V }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, ({ value }, ..._: I) => {
      return schema.safeParse(value);
    }) as Procedure<T, I, V>;
  };

  procedure.zodAsync = <V>(schema: { parseAsync: (data: any) => V }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, ({ value }, ..._: I) => {
      return schema.parseAsync(value);
    }) as Procedure<T, I, V>;
  };

  procedure.zodSafeAsync = <V>(schema: {
    safeParseAsync: (data: any) => V;
  }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, ({ value }, ..._: I) => {
      return schema.safeParseAsync(value);
    }) as Procedure<T, I, V>;
  };

  return procedure as Procedure<T, I, O>;
}

export function createContext<T extends Record<string, any>>(
  context: () => T,
  options: Partial<ContextOptions> = {}
) {
  const { immidiate, ...procOptions } = options;

  // call context right away
  if (immidiate === true) {
    const ctx = context();
    context = () => ctx;
  }

  const ctx = () => context();
  ctx.context = context;
  ctx.procedure = <I extends any[], O>(fn: (context: T, ...input: I) => O) => {
    return createProcedure<T, I, O>(context, fn, procOptions);
  };

  return ctx;
}

export function createPipe<I extends any[], O>(
  fn: (...input: I) => O,
  options: Partial<ProcedureOptions> = {}
) {
  // recursive pipe generator
  const generatePipe = <Vprev, Vnext>(
    injection: (...input: I) => Vprev,
    pipe: (value: Vprev, ...input: I) => Vnext
  ): Pipes<I, Vnext> => {
    // pipe injector for context and previous value
    const injector = (...input: I) => {
      const value = injection(...input);
      return pipe(value, ...input);
    };

    // create pipe caller
    const caller = (...input: I) => {
      const value = injection(...input);
      const final = pipe(value, ...input);

      if (typeof options.onAfterCalled === 'function') {
        options.onAfterCalled(final);
      }

      return final;
    };

    // generate pipe function recursively
    caller.pipe = <V>(pipeFn: (context: Vnext, ...input: I) => V) => {
      return generatePipe<Vnext, V>(injector, pipeFn) as Pipes<I, V>;
    };

    // create infer function
    caller.infer = <V>(inferFn: (inferer: Inferer) => V) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, (value, ..._: I) => {
        return inferFn(infer(value as DataType));
      }) as Pipes<I, V>;
    };

    caller.zod = <V>(schema: { parse: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, (value, ..._: I) => {
        return schema.parse(value);
      }) as Pipes<I, V>;
    };

    caller.zodSafe = <V>(schema: { safeParse: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, (value, ..._: I) => {
        return schema.safeParse(value);
      }) as Pipes<I, V>;
    };

    caller.zodAsync = <V>(schema: { parseAsync: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, (value, ..._: I) => {
        return schema.parseAsync(value);
      }) as Pipes<I, V>;
    };

    caller.zodSafeAsync = <V>(schema: { safeParseAsync: (data: any) => V }) => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      return generatePipe<Vnext, V>(injector, (value, ..._: I) => {
        return schema.safeParseAsync(value);
      }) as Pipes<I, V>;
    };

    // return as new procedure
    return caller as Pipes<I, Vnext>;
  };

  // prepare caller
  const pipes = (...input: I) => {
    if (typeof options.onBeforeCalled === 'function') {
      options.onBeforeCalled(input);
    }

    const value = fn(...input);

    if (typeof options.onAfterCalled === 'function') {
      options.onAfterCalled(value);
    }

    return value;
  };

  // create initial injector
  const injector = (...input: I) => {
    if (typeof options.onBeforeCalled === 'function') {
      options.onBeforeCalled(input);
    }

    return fn(...input);
  };

  // create first pipe function
  pipes.pipe = <V>(pipeFn: (value: O, ...input: I) => V) => {
    return generatePipe<O, V>(injector, pipeFn) as Pipes<I, V>;
  };

  // create first infer function
  pipes.infer = <V>(inferFn: (inferer: Inferer) => V) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, (value, ..._: I) => {
      return inferFn(infer(value as DataType));
    }) as Pipes<I, V>;
  };

  pipes.zod = <V>(schema: { parse: (data: any) => V }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, (value, ..._: I) => {
      return schema.parse(value);
    }) as Pipes<I, V>;
  };

  pipes.zodSafe = <V>(schema: { safeParse: (data: any) => V }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, (value, ..._: I) => {
      return schema.safeParse(value);
    }) as Pipes<I, V>;
  };

  pipes.zodAsync = <V>(schema: { parseAsync: (data: any) => V }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, (value, ..._: I) => {
      return schema.parseAsync(value);
    }) as Pipes<I, V>;
  };

  pipes.zodSafeAsync = <V>(schema: { safeParseAsync: (data: any) => V }) => {
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    return generatePipe<O, V>(injector, (value, ..._: I) => {
      return schema.safeParseAsync(value);
    }) as Pipes<I, V>;
  };

  return pipes as Pipes<I, O>;
}
