/* eslint-disable @typescript-eslint/no-explicit-any */
import { AccessSchema, Schema } from '../permit';
import {
  Context,
  ContextValue,
  GuardFunction,
  Pipe,
  ProcedureError,
  onCall,
  onError,
} from './types';

export class ProcedureBuilder<T> {
  private _context: () => Context<T>;
  private _onErrorHook: onError;
  private _onCallHook: onCall<T>;

  constructor(context: () => Context<T>) {
    this._context = context;
    this._onErrorHook = (error) => error;
    this._onCallHook = () => undefined;
  }

  set context(context: () => Context<T>) {
    this._context = context;
  }

  get context() {
    return this._context;
  }

  set onError(hook: onError) {
    this._onErrorHook = hook;
  }

  set onCall(hook: onCall<T>) {
    this._onCallHook = hook;
  }

  build<I extends any[], V>(
    key: string,
    fn: (context: Context<T>, ...input: I) => V
  ) {
    return this.recursivePipeGenerator<I, null, V>(fn, () => {
      const context = this._context();
      return { context, value: null, path: key };
    });
  }

  private throwError(path: string, message: string, error: Error) {
    const err = new ProcedureError(path, message, error);
    this._onErrorHook(err);

    return err;
  }

  private recursivePipeGenerator<I extends any[], V, VNext>(
    func: (context: ContextValue<T, V>, ...input: I) => VNext,
    endFn: (...input: I) => {
      value: V;
      context: Context<T>;
      guard?: (scopes: string[]) => boolean;
      path: string;
    }
  ) {
    // prepare recursive pipe function
    const pipe = <Output>(
      fn: (context: ContextValue<T, VNext>, ...input: I) => Output
    ) => {
      return this.recursivePipeGenerator<I, VNext, Output>(
        fn,
        (...input: I) => {
          const { context, value, guard, path } = endFn(...input);
          return {
            guard,
            context,
            path: path + '.pipe',
            value: func({ ...context, value }, ...input),
          };
        }
      );
    };

    // prepare recursive permit function
    const permit = (scopes: string[]) => {
      return this.recursivePipeGenerator<I, V, VNext>(func, (...input: I) => {
        const { context, value, guard, path } = endFn(...input);
        const old = context.scopes ?? [];
        if (typeof guard === 'function') {
          context.protected = guard([...new Set([...old, ...scopes])]);
        }

        console.log('permit', {
          path: path + '.permit',
          scopes: [...new Set([...old, ...scopes])],
        });
        return { context, value, guard, path: path + '.permit' };
      });
    };

    // perpare recursive protect function
    const protect = <S extends Schema>(
      schema: AccessSchema<S>,
      guard: GuardFunction<S>
    ) => {
      return this.recursivePipeGenerator<I, V, VNext>(func, (...input: I) => {
        const { context, value, guard: prevGuard, path } = endFn(...input);
        const scopes = context.scopes ?? [];

        const extendGuard = (scopes: string[]) => {
          const permit = schema.permit(scopes);
          if (typeof prevGuard === 'function') {
            const prev = prevGuard(scopes);
            const curr = guard(permit);
            console.log('protect', { prev, curr });
            return prev && curr;
          } else {
            return guard(permit);
          }
        };

        if (scopes.length !== 0) {
          context.protected = !!context.protected && extendGuard(scopes);
        } else if (context.protected === undefined) {
          context.protected = false;
        }

        return { context, value, guard: extendGuard, path: path + '.protect' };
      });
    };

    const onCallHook = this._onCallHook;
    const onErrorHook = this._onErrorHook;
    const throwError = this.throwError;
    return {
      pipe,
      permit,
      protect,
      get fn() {
        return (...input: I) => {
          const { context, value, path } = endFn(...input);
          const endPath = path + '.fn';
          try {
            const contextValue = { ...context, value };
            onCallHook(endPath, contextValue, input);
            return func(contextValue, ...input);
          } catch (error) {
            if (typeof error === 'string') {
              throw throwError(endPath, error, new Error(error));
            } else if (error instanceof ProcedureError) {
              error.path = endPath;
              throw onErrorHook(error);
            } else if (error instanceof Error) {
              throw throwError(endPath, error.message, error);
            }
          }

          const err = new Error();
          throw throwError(endPath, 'Unkown Procedure Error', err);
        };
      },
    } as Pipe<T, I, VNext>;
  }
}
