import { describe, expect, test } from 'vitest';
import { createContext, createPipe, createProcedure } from '..';
import { z } from 'zod';

// import { createSchema } from '../../permit';
// import { z } from 'zod';

// const schema = {
//   secret: {
//     read: 'read value of secret',
//     write: 'update value of secret',
//   },
// };

// const access = createSchema(schema);

describe('Garda Unit Testing: Public Context', () => {
  const ctx = () => {
    let count = 0;
    return {
      add() {
        count++;
      },
      get count() {
        return count;
      },
      set count(value: number) {
        count = value;
      },
    };
  };

  /**
   * Test context that closed between call.
   * in every call, count should reset to 0.
   */
  test('closed context test', () => {
    const context = createContext(ctx);
    const getCount = context.procedure((ctx) => ctx.count);
    const counter = context.procedure((ctx) => {
      ctx.add();
      return ctx.count;
    });
    const doubleCounter = context.procedure((ctx) => {
      ctx.count = ctx.count + 2;
      return ctx.count;
    });

    const customCounter = context.procedure((ctx, num: number) => {
      ctx.count = ctx.count + num;
      return ctx.count;
    });

    // test get count must be still 0
    expect(getCount()).toBe(0);

    // test add count, but this should still be 0
    expect(counter()).toBe(1);
    expect(getCount()).toBe(0);

    expect(doubleCounter()).toBe(2);
    expect(getCount()).toBe(0);

    expect(customCounter(20)).toBe(20);
    expect(getCount()).toBe(0);
  });

  /**
   * Test context that shared between call.
   * count changed after call
   */
  test('shared context test', () => {
    const context = createContext(ctx, { immidiate: true });
    const getCount = context.procedure((ctx) => ctx.count);
    const counter = context.procedure((ctx) => {
      ctx.add();
      return ctx.count;
    });
    const doubleCounter = context.procedure((ctx) => {
      ctx.count = ctx.count + 2;
      return ctx.count;
    });

    const customCounter = context.procedure((ctx, num: number) => {
      ctx.count = ctx.count + num;
      return ctx.count;
    });

    expect(getCount()).toBe(0);

    expect(counter()).toBe(1);
    expect(getCount()).toBe(1);

    expect(doubleCounter()).toBe(3);
    expect(getCount()).toBe(3);

    expect(customCounter(20)).toBe(23);
    expect(getCount()).toBe(23);
  });

  /**
   * Pipe function with closed context should still carry
   * calculated value from previous pipe/procedure
   */
  test('pipe function utility test', () => {
    const context = createContext(ctx);
    const count = context.procedure((ctx) => ctx.count);
    const addCountBy2 = count.pipe((ctx) => ctx.value + 2);
    const subtractCount = (num: number) => {
      const calc = addCountBy2.pipe((ctx) => ctx.value - num);
      return calc();
    };

    expect(count()).toBe(0);
    expect(addCountBy2()).toBe(2);
    expect(subtractCount(1)).toBe(1);

    // when count called, count should still be 0
    expect(count()).toBe(0);
  });

  /** Infer chain function shoul work properly */
  test('infer function utility test', () => {
    const context = () => ({
      user: JSON.stringify({
        id: 'c8244705-a479-4af4-914c-e30574f2c7e6',
        firstName: 'Elmo',
        lastName: 'Rosenbaum',
      }),
    });

    const getUserName = createProcedure(context, (ctx) =>
      JSON.parse(ctx.user)
    ).infer((i) => `${i.string('firstName')} ${i.string('lastName')}`);

    expect(getUserName()).toBe('Elmo Rosenbaum');
  });

  /** Zod chain function should work properly */
  test('zod function utility test', () => {
    const context = createContext(() => ({
      user: JSON.stringify({
        id: 'c8244705-a479-4af4-914c-e30574f2c7e6',
        firstName: 'Elmo',
        lastName: 'Rosenbaum',
      }),
    }));

    const userSchema = z.object({
      id: z.string(),
      firstName: z.string(),
      lastName: z.string(),
    });

    const zodParse = context
      .procedure((ctx) => JSON.parse(ctx.user))
      .zod(userSchema);
    const zodSafeParse = context
      .procedure((ctx) => JSON.parse(ctx.user))
      .zodSafe(userSchema);

    expect(zodParse().firstName).toBe('Elmo');
    expect(zodSafeParse().success).toBeTruthy();
  });

  /** Error throw testing  */
  test('on error handling', () => {
    const context = createContext(ctx);
    const onThrowError = context.procedure(() => {
      throw new Error('Error Throwed');
    });

    expect(onThrowError).toThrowError('Error Throwed');

    const deepPipeThrow = context
      .procedure(() => 1)
      .pipe(() => 2)
      .pipe(() => 3)
      .pipe(() => 4)
      .pipe(() => 5)
      .pipe(() => 6)
      .pipe(() => 7)
      .pipe(() => {
        throw new Error('8');
      });

    expect(deepPipeThrow).toThrowError('8');
  });

  /**
   * Non context procedure test
   */
  test('pipes (non-context procedure) function utility', () => {
    const fn = createPipe((num: number) => num)
      .pipe((count, num) => count + num)
      .pipe((count, num) => count + num)
      .pipe((count, num) => count + num)
      .pipe((count, num) => count + num);

    expect(fn(1)).toBe(5);
  });

  /** Sequence work as expected */
  test('call sequence on pipe', () => {
    const runner = createProcedure(
      () => ({}),
      () => ['procedure']
    )
      .pipe(({ value }) => {
        value.push('pipe 1');
        return value;
      })
      .pipe(({ value }) => {
        value.push('pipe 2');
        return value;
      })
      .pipe(({ value }) => {
        value.push('pipe 3');
        return value;
      })
      .pipe(({ value }) => {
        value.push('joiner');
        return value.join(', ');
      });

    expect(runner()).toBe('procedure, pipe 1, pipe 2, pipe 3, joiner');

    const pipes = createPipe(() => ['function'])
      .pipe((value) => {
        value.push('pipe 1');
        return value;
      })
      .pipe((value) => {
        value.push('pipe 2');
        return value;
      })
      .pipe((value) => {
        value.push('pipe 3');
        return value;
      })
      .pipe((value) => {
        value.push('joiner');
        return value.join(', ');
      });

    expect(pipes()).toBe('function, pipe 1, pipe 2, pipe 3, joiner');
  });

  /** Hook Utility test */
  test('onCalled hook utility test', () => {
    const afterContext = createContext(() => ({}), {
      onAfterCalled(value) {
        throw new Error(`value: ${value}`);
      },
    });

    const afterHook = afterContext.procedure(() => 'procedure');
    expect(afterHook).toThrowError('value: procedure');
    expect(afterHook.pipe(() => 'pipe 1')).toThrowError('value: pipe 1');
    expect(afterHook.pipe(() => 1).pipe(() => 2)).toThrowError('value: 2');

    const beforeContext = createContext(() => ({}), {
      onBeforeCalled(input) {
        throw new Error(`input: ${(input ?? []).join(', ')}`);
      },
    });

    const beforeHook = beforeContext.procedure(
      (_, a: number, b: string) => a + b
    );
    expect(() => beforeHook(2, 'dua')).toThrowError('input: 2, dua');
    expect(() => beforeHook(1, 'satu')).toThrowError('input: 1, satu');
  });
});
